@@ -71,7 +71,6 @@ static int apfs_create_snap_metadata_rec(struct inode *mntpoint, struct apfs_nod
 	struct apfs_sb_info *sbi = APFS_SB(sb);
 	struct apfs_superblock *vsb_raw = sbi->s_vsb_raw;
 	struct apfs_query *query = NULL;
-	struct apfs_key key = {0};
 	struct apfs_snap_metadata_key raw_key;
 	struct apfs_snap_metadata_val *raw_val = NULL;
 	int val_len;
@@ -79,14 +78,12 @@ static int apfs_create_snap_metadata_rec(struct inode *mntpoint, struct apfs_nod
 	u64 xid = APFS_NXI(sb)->nx_xid;
 	int err;
 
-	apfs_init_snap_metadata_key(xid, &key);
-
 	query = apfs_alloc_query(snap_root, NULL /* parent */);
 	if (!query) {
 		err = -ENOMEM;
 		goto fail;
 	}
-	query->key = &key;
+	apfs_init_snap_metadata_key(xid, &query->key);
 	query->flags |= APFS_QUERY_SNAP_META | APFS_QUERY_EXACT;
 
 	err = apfs_btree_query(sb, &query);
@@ -116,7 +113,7 @@ static int apfs_create_snap_metadata_rec(struct inode *mntpoint, struct apfs_nod
 	raw_val->extentref_tree_type = vsb_raw->apfs_extentref_tree_type;
 	raw_val->flags = 0;
 	raw_val->name_len = cpu_to_le16(name_len + 1); /* Count the null byte */
-	strcpy(raw_val->name, name);
+	strscpy(raw_val->name, name, name_len + 1);
 
 	apfs_assert_in_transaction(sb, &vsb_raw->apfs_o);
 	raw_val->inum = vsb_raw->apfs_next_obj_id;
@@ -137,27 +134,24 @@ static int apfs_create_snap_name_rec(struct apfs_node *snap_root, const char *na
 {
 	struct super_block *sb = snap_root->object.sb;
 	struct apfs_query *query = NULL;
-	struct apfs_key key = {0};
 	struct apfs_snap_name_key *raw_key = NULL;
 	struct apfs_snap_name_val raw_val;
 	int key_len;
 	int err;
 
-	apfs_init_snap_name_key(name, &key);
-
 	query = apfs_alloc_query(snap_root, NULL /* parent */);
 	if (!query) {
 		err = -ENOMEM;
 		goto fail;
 	}
-	query->key = &key;
+	apfs_init_snap_name_key(name, &query->key);
 	query->flags |= APFS_QUERY_SNAP_META | APFS_QUERY_EXACT;
 
 	err = apfs_btree_query(sb, &query);
 	if (err == 0) {
-		/* TODO: avoid transaction abort here */
-		apfs_info(sb, "a snapshot with that name already exists");
-		err = -EEXIST;
+		/* This should never happen here, the caller has checked */
+		apfs_alert(sb, "snapshot name record already exists");
+		err = -EFSCORRUPTED;
 		goto fail;
 	}
 	if (err != -ENODATA) {
@@ -173,7 +167,7 @@ static int apfs_create_snap_name_rec(struct apfs_node *snap_root, const char *na
 	}
 	apfs_key_set_hdr(APFS_TYPE_SNAP_NAME, APFS_SNAP_NAME_OBJ_ID, raw_key);
 	raw_key->name_len = cpu_to_le16(name_len + 1); /* Count the null byte */
-	strcpy(raw_key->name, name);
+	strscpy(raw_key->name, name, name_len + 1);
 
 	raw_val.snap_xid = cpu_to_le64(APFS_NXI(sb)->nx_xid);
 
@@ -246,7 +240,6 @@ static int apfs_update_omap_snap_tree(struct super_block *sb, __le64 *oid_p)
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
 	struct apfs_node *root = NULL;
 	u64 oid = le64_to_cpup(oid_p);
-	struct apfs_key key = {0};
 	struct apfs_query *query = NULL;
 	__le64 raw_key;
 	struct apfs_omap_snapshot raw_val = {0};
@@ -268,14 +261,12 @@ static int apfs_update_omap_snap_tree(struct super_block *sb, __le64 *oid_p)
 	}
 	oid = 0;
 
-	apfs_init_omap_snap_key(nxi->nx_xid, &key);
-
 	query = apfs_alloc_query(root, NULL /* parent */);
 	if (!query) {
 		err = -ENOMEM;
 		goto fail;
 	}
-	query->key = &key;
+	apfs_init_omap_snap_key(nxi->nx_xid, &query->key);
 	query->flags = APFS_QUERY_OMAP_SNAP | APFS_QUERY_EXACT;
 
 	err = apfs_btree_query(sb, &query);
@@ -341,6 +332,56 @@ static int apfs_update_omap_snapshots(struct super_block *sb)
 	return err;
 }
 
+/**
+ * apfs_snapshot_name_check - Check if a snapshot with the given name exists
+ * @sb:		filesystem superblock
+ * @name:	name to check
+ * @name_len:	length of @name
+ * @eexist:	on return, whether the name exists or not
+ *
+ * Returns 0 on success, or a negative error code in case of failure.
+ */
+static int apfs_snapshot_name_check(struct super_block *sb, const char *name, int name_len, bool *eexist)
+{
+	struct apfs_sb_info *sbi = APFS_SB(sb);
+	struct apfs_superblock *vsb_raw = sbi->s_vsb_raw;
+	struct apfs_node *snap_root = NULL;
+	struct apfs_query *query = NULL;
+	int err;
+
+	snap_root = apfs_read_node(sb, le64_to_cpu(vsb_raw->apfs_snap_meta_tree_oid), APFS_OBJ_PHYSICAL, false /* write */);
+	if (IS_ERR(snap_root)) {
+		apfs_err(sb, "failed to read snap meta root 0x%llx", le64_to_cpu(vsb_raw->apfs_snap_meta_tree_oid));
+		return PTR_ERR(snap_root);
+	}
+	vsb_raw = NULL;
+
+	query = apfs_alloc_query(snap_root, NULL /* parent */);
+	if (!query) {
+		err = -ENOMEM;
+		goto out;
+	}
+	apfs_init_snap_name_key(name, &query->key);
+	query->flags |= APFS_QUERY_SNAP_META | APFS_QUERY_EXACT;
+
+	err = apfs_btree_query(sb, &query);
+	if (err == 0) {
+		*eexist = true;
+	} else if (err == -ENODATA) {
+		*eexist = false;
+		err = 0;
+	} else {
+		apfs_err(sb, "query failed (%s)", name);
+	}
+
+out:
+	apfs_free_query(query);
+	query = NULL;
+	apfs_node_free(snap_root);
+	snap_root = NULL;
+	return err;
+}
+
 /**
  * apfs_do_ioc_takesnapshot - Actual work for apfs_ioc_take_snapshot()
  * @mntpoint:	inode of the mount point to snapshot
@@ -357,12 +398,31 @@ static int apfs_do_ioc_take_snapshot(struct inode *mntpoint, const char *name, i
 	/* TODO: remember to update the maxops in the future */
 	struct apfs_max_ops maxops = {0};
 	u64 sblock_oid;
+	bool eexist = false;
 	int err;
 
 	err = apfs_transaction_start(sb, maxops);
 	if (err)
 		return err;
 
+	/*
+	 * This check can fail in normal operation, so run it before making any
+	 * changes and exit the transaction cleanly if necessary. The name
+	 * lookup will have to be repeated later, but it's ok because I don't
+	 * expect snapshot creation to be a bottleneck for anyone.
+	 */
+	err = apfs_snapshot_name_check(sb, name, name_len, &eexist);
+	if (err) {
+		apfs_err(sb, "snapshot name existence check failed");
+		goto fail;
+	}
+	if (eexist) {
+		err = apfs_transaction_commit(sb);
+		if (err)
+			goto fail;
+		return -EEXIST;
+	}
+
 	/*
 	 * Flush the extent caches to the extenref tree before it gets moved to
 	 * the snapshot. It seems safer in general to avoid big unpredictable
@@ -499,15 +559,12 @@ static int apfs_snapshot_name_to_xid(struct apfs_node *snap_root, const char *na
 {
 	struct super_block *sb = snap_root->object.sb;
 	struct apfs_query *query = NULL;
-	struct apfs_key key = {0};
 	int err;
 
-	apfs_init_snap_name_key(name, &key);
-
 	query = apfs_alloc_query(snap_root, NULL /* parent */);
 	if (!query)
 		return -ENOMEM;
-	query->key = &key;
+	apfs_init_snap_name_key(name, &query->key);
 	query->flags |= APFS_QUERY_SNAP_META | APFS_QUERY_EXACT;
 
 	err = apfs_btree_query(sb, &query);
@@ -545,15 +602,12 @@ static int apfs_snapshot_xid_to_sblock(struct apfs_node *snap_root, u64 xid, u64
 {
 	struct super_block *sb = snap_root->object.sb;
 	struct apfs_query *query = NULL;
-	struct apfs_key key = {0};
 	int err;
 
-	apfs_init_snap_metadata_key(xid, &key);
-
 	query = apfs_alloc_query(snap_root, NULL /* parent */);
 	if (!query)
 		return -ENOMEM;
-	query->key = &key;
+	apfs_init_snap_metadata_key(xid, &query->key);
 	query->flags |= APFS_QUERY_SNAP_META | APFS_QUERY_EXACT;
 
 	err = apfs_btree_query(sb, &query);
