@@ -7,354 +7,6 @@
 #include <linux/rmap.h>
 #include "apfs.h"
 
-/**
- * apfs_cpoint_init_area - Initialize the new blocks of a checkpoint area
- * @sb:		superblock structure
- * @base:	first block of the area
- * @blks:	block count for the area
- * @next:	first block for the new checkpoint
- * @len:	block count for the new checkpoint
- *
- * Copies the blocks belonging to the last checkpoint to the area assigned for
- * the new one, updating their xids, oids and checksums.  Returns 0 on success,
- * or a negative error code in case of failure.
- */
-static int apfs_cpoint_init_area(struct super_block *sb, u64 base, u32 blks,
-				 u32 next, u32 len)
-{
-	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	u32 i;
-
-	for (i = 0; i < len; ++i) {
-		struct buffer_head *new_bh, *old_bh;
-		struct apfs_obj_phys *new_obj;
-		u32 new_index = (next + i) % blks;
-		u32 old_index = (blks + new_index - len) % blks;
-		u32 type;
-		int err;
-
-		new_bh = apfs_getblk(sb, base + new_index);
-		old_bh = apfs_sb_bread(sb, base + old_index);
-		if (!new_bh || !old_bh) {
-			apfs_err(sb, "unable to read the checkpoint areas");
-			brelse(new_bh);
-			brelse(old_bh);
-			return -EINVAL;
-		}
-		memcpy(new_bh->b_data, old_bh->b_data, sb->s_blocksize);
-		brelse(old_bh);
-
-		if (nxi->nx_flags & APFS_CHECK_NODES && !apfs_obj_verify_csum(sb, new_bh)) {
-			apfs_err(sb, "bad checksum for checkpoint area block");
-			brelse(new_bh);
-			return -EFSBADCRC;
-		}
-
-		new_obj = (struct apfs_obj_phys *)new_bh->b_data;
-		type = le32_to_cpu(new_obj->o_type);
-		if ((type & APFS_OBJ_STORAGETYPE_MASK) == APFS_OBJ_PHYSICAL)
-			new_obj->o_oid = cpu_to_le64(new_bh->b_blocknr);
-
-		ASSERT(nxi->nx_xid == le64_to_cpu(new_obj->o_xid) + 1);
-		new_obj->o_xid = cpu_to_le64(nxi->nx_xid);
-		err = apfs_transaction_join(sb, new_bh);
-		if (err) {
-			brelse(new_bh);
-			return err;
-		}
-
-		/* The transaction isn't over, don't commit the superblock */
-		if ((type & APFS_OBJECT_TYPE_MASK) !=
-		    APFS_OBJECT_TYPE_NX_SUPERBLOCK)
-			set_buffer_csum(new_bh);
-
-		new_obj = NULL;
-		brelse(new_bh);
-	}
-	return 0;
-}
-
-/**
- * apfs_cpoint_init_desc - Initialize the descriptor area for a new checkpoint
- * @sb: superblock structure
- *
- * Copies the latest descriptor and mapping blocks into the new checkpoint;
- * updates both the on-disk and in-memory superblocks to refer to the new
- * transaction.  Returns 0 on success, or a negative error code in case of
- * failure.
- */
-static int apfs_cpoint_init_desc(struct super_block *sb)
-{
-	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
-	struct buffer_head *new_sb_bh = NULL;
-	u64 desc_base = le64_to_cpu(raw_sb->nx_xp_desc_base);
-	u32 desc_next = le32_to_cpu(raw_sb->nx_xp_desc_next);
-	u32 desc_blks = le32_to_cpu(raw_sb->nx_xp_desc_blocks);
-	u32 desc_len = le32_to_cpu(raw_sb->nx_xp_desc_len);
-	u32 new_sb_index;
-	int err;
-
-	if (!desc_blks || !desc_len) {
-		apfs_err(sb, "bad checkpoint descriptor area");
-		return -EFSCORRUPTED;
-	}
-
-	err = apfs_cpoint_init_area(sb, desc_base, desc_blks,
-				    desc_next, desc_len);
-	if (err) {
-		apfs_err(sb, "failed to init area");
-		return err;
-	}
-
-	/* Now update the superblock with the new checkpoint */
-
-	new_sb_index = (desc_next + desc_len - 1) % desc_blks;
-	new_sb_bh = apfs_sb_bread(sb, desc_base + new_sb_index);
-	if (!new_sb_bh) {
-		apfs_err(sb, "unable to read the new checkpoint superblock");
-		brelse(new_sb_bh);
-		return -EINVAL;
-	}
-
-	brelse(nxi->nx_object.o_bh);
-	nxi->nx_object.o_bh = new_sb_bh;
-	nxi->nx_object.data = new_sb_bh->b_data;
-	raw_sb = (struct apfs_nx_superblock *)new_sb_bh->b_data;
-	nxi->nx_raw = raw_sb;
-	nxi->nx_object.block_nr = new_sb_bh->b_blocknr;
-
-	ASSERT(nxi->nx_xid == le64_to_cpu(raw_sb->nx_next_xid));
-	ASSERT(buffer_trans(new_sb_bh));
-	raw_sb->nx_next_xid = cpu_to_le64(nxi->nx_xid + 1);
-
-	/* Apparently the previous checkpoint gets invalidated right away */
-	raw_sb->nx_xp_desc_index = cpu_to_le32(desc_next);
-	desc_next = (new_sb_index + 1) % desc_blks;
-	raw_sb->nx_xp_desc_next = cpu_to_le32(desc_next);
-	return 0;
-}
-
-/**
- * apfs_cpoint_init_data - Initialize the data area for a new checkpoint
- * @sb: superblock structure
- *
- * Copies the latest data blocks into the new checkpoint, updating the fields
- * related to this change of location.  Returns 0 on success, or a negative
- * error code in case of failure.
- */
-static int apfs_cpoint_init_data(struct super_block *sb)
-{
-	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
-	u64 data_base = le64_to_cpu(raw_sb->nx_xp_data_base);
-	u32 data_next = le32_to_cpu(raw_sb->nx_xp_data_next);
-	u32 data_blks = le32_to_cpu(raw_sb->nx_xp_data_blocks);
-	u32 data_len = le32_to_cpu(raw_sb->nx_xp_data_len);
-	int err;
-
-	if (!data_blks || !data_len) {
-		apfs_err(sb, "bad checkpoint data area");
-		return -EFSCORRUPTED;
-	}
-
-	err = apfs_cpoint_init_area(sb, data_base, data_blks,
-				    data_next, data_len);
-	if (err) {
-		apfs_err(sb, "failed to init area");
-		return err;
-	}
-
-	/* Apparently the previous checkpoint gets invalidated right away */
-	apfs_assert_in_transaction(sb, &raw_sb->nx_o);
-	raw_sb->nx_xp_data_index = cpu_to_le32(data_next);
-	data_next = (data_next + data_len) % data_blks;
-	raw_sb->nx_xp_data_next = cpu_to_le32(data_next);
-	return 0;
-}
-
-/**
- * apfs_update_mapping - Update a single checkpoint mapping
- * @sb:		superblock structure
- * @map:	checkpoint mapping to update
- *
- * Remaps @map to point to the address of the same ephemeral object in the next
- * checkpoint data area.
- */
-static void apfs_update_mapping(struct super_block *sb,
-				struct apfs_checkpoint_mapping *map)
-{
-	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
-	u64 data_base = le64_to_cpu(raw_sb->nx_xp_data_base);
-	u32 data_blks = le32_to_cpu(raw_sb->nx_xp_data_blocks);
-	u32 data_len = le32_to_cpu(raw_sb->nx_xp_data_len);
-	u64 old_paddr, new_paddr;
-	u32 old_index, new_index;
-
-	old_paddr = le64_to_cpu(map->cpm_paddr);
-	old_index = old_paddr - data_base;
-	new_index = (old_index + data_len) % data_blks;
-	new_paddr = data_base + new_index;
-
-	map->cpm_paddr = cpu_to_le64(new_paddr);
-}
-
-/**
- * apfs_update_mapping_blocks - Update all checkpoint mappings
- * @sb: superblock structure
- *
- * Updates the mappings in the checkpoint descriptor area to point to the
- * address of the ephemeral objects for the new checkpoint.  Returns 0 on
- * success, or a negative error code in case of failure.
- */
-static int apfs_update_mapping_blocks(struct super_block *sb)
-{
-	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
-	u64 desc_base = le64_to_cpu(raw_sb->nx_xp_desc_base);
-	u32 desc_index = le32_to_cpu(raw_sb->nx_xp_desc_index);
-	u32 desc_blks = le32_to_cpu(raw_sb->nx_xp_desc_blocks);
-	u32 desc_len = le32_to_cpu(raw_sb->nx_xp_desc_len);
-	u32 i;
-
-	/* Last block in the area is superblock; the rest are mapping blocks */
-	for (i = 0; i < desc_len - 1; ++i) {
-		struct buffer_head *bh;
-		struct apfs_checkpoint_map_phys *cpm;
-		u32 desc_curr = (desc_index + i) % desc_blks;
-		u32 map_count;
-		int j;
-
-		bh = apfs_sb_bread(sb, desc_base + desc_curr);
-		if (!bh) {
-			apfs_err(sb, "failed to read cpm block");
-			return -EINVAL;
-		}
-		ASSERT(buffer_trans(bh));
-		cpm = (struct apfs_checkpoint_map_phys *)bh->b_data;
-		apfs_assert_in_transaction(sb, &cpm->cpm_o);
-
-		map_count = le32_to_cpu(cpm->cpm_count);
-		if (map_count > apfs_max_maps_per_block(sb)) {
-			apfs_err(sb, "block has too many maps (%d)", map_count);
-			brelse(bh);
-			return -EFSCORRUPTED;
-		}
-
-		for (j = 0; j < map_count; ++j)
-			apfs_update_mapping(sb, &cpm->cpm_map[j]);
-		set_buffer_csum(bh);
-		brelse(bh);
-	}
-	return 0;
-}
-
-/**
- * apfs_cpoint_data_allocate - Allocate a new block in the checkpoint data area
- * @sb:		superblock structure
- * @bno:	on return, the allocated block number
- */
-void apfs_cpoint_data_allocate(struct super_block *sb, u64 *bno)
-{
-	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
-	u64 data_base = le64_to_cpu(raw_sb->nx_xp_data_base);
-	u32 data_next = le32_to_cpu(raw_sb->nx_xp_data_next);
-	u32 data_blks = le32_to_cpu(raw_sb->nx_xp_data_blocks);
-	u32 data_len = le32_to_cpu(raw_sb->nx_xp_data_len);
-
-	*bno = data_base + data_next;
-	data_next = (data_next + 1) % data_blks;
-	data_len++;
-
-	apfs_assert_in_transaction(sb, &raw_sb->nx_o);
-	raw_sb->nx_xp_data_next = cpu_to_le32(data_next);
-	raw_sb->nx_xp_data_len = cpu_to_le32(data_len);
-}
-
-/**
- * apfs_cpoint_data_free - Free a block in the checkpoint data area
- * @sb:		superblock structure
- * @bno:	block number to free
- *
- * Returns 0 on sucess, or a negative error code in case of failure.
- */
-int apfs_cpoint_data_free(struct super_block *sb, u64 bno)
-{
-	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
-	u32 data_next = le32_to_cpu(raw_sb->nx_xp_data_next);
-	u32 data_blks = le32_to_cpu(raw_sb->nx_xp_data_blocks);
-	u32 data_len = le32_to_cpu(raw_sb->nx_xp_data_len);
-	u32 i, bno_i;
-
-	/*
-	 * We can't leave a hole in the data area, so we need to shift all
-	 * blocks that come after @bno one position back.
-	 */
-	bno_i = apfs_index_in_data_area(sb, bno);
-	for (i = bno_i; i < data_len - 1; ++i) {
-		struct buffer_head *old_bh, *new_bh;
-		int err;
-
-		new_bh = apfs_getblk(sb, apfs_data_index_to_bno(sb, i));
-		old_bh = apfs_sb_bread(sb, apfs_data_index_to_bno(sb, i + 1));
-		if (!new_bh || !old_bh) {
-			apfs_err(sb, "failed to map blocks in data area");
-			brelse(new_bh);
-			brelse(old_bh);
-			return -EIO;
-		}
-		/* I could also just remap the buffer heads... */
-		memcpy(new_bh->b_data, old_bh->b_data, sb->s_blocksize);
-
-		brelse(old_bh);
-		err = apfs_transaction_join(sb, new_bh); /* Not really needed */
-		brelse(new_bh);
-		if (err)
-			return err;
-	}
-
-	data_next = (data_blks + data_next - 1) % data_blks;
-	data_len--;
-
-	apfs_assert_in_transaction(sb, &raw_sb->nx_o);
-	raw_sb->nx_xp_data_next = cpu_to_le32(data_next);
-	raw_sb->nx_xp_data_len = cpu_to_le32(data_len);
-	return 0;
-}
-
-/**
- * apfs_checkpoint_start - Start the checkpoint for a new transaction
- * @sb:	superblock structure
- *
- * Sets the descriptor and data areas for a new checkpoint.  Returns 0 on
- * success, or a negative error code in case of failure.
- */
-static int apfs_checkpoint_start(struct super_block *sb)
-{
-	int err;
-
-	ASSERT(!(sb->s_flags & SB_RDONLY));
-
-	err = apfs_cpoint_init_desc(sb);
-	if (err) {
-		apfs_err(sb, "failed to init descriptor area");
-		return err;
-	}
-	err = apfs_cpoint_init_data(sb);
-	if (err) {
-		apfs_err(sb, "failed to init data area");
-		return err;
-	}
-	err = apfs_update_mapping_blocks(sb);
-	if (err)
-		apfs_err(sb, "failed to update checkpoint mappings");
-
-	return err;
-}
-
 /**
  * apfs_checkpoint_end - End the new checkpoint
  * @sb:	filesystem superblock
@@ -367,24 +19,40 @@ static int apfs_checkpoint_end(struct super_block *sb)
 {
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
 	struct apfs_obj_phys *obj = &nxi->nx_raw->nx_o;
-	struct buffer_head *bh = nxi->nx_object.o_bh;
+	struct buffer_head *bh = NULL;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 10, 0)
+	struct address_space *bdev_map = nxi->nx_bdev->bd_mapping;
+#else
 	struct inode *bdev_inode = nxi->nx_bdev->bd_inode;
 	struct address_space *bdev_map = bdev_inode->i_mapping;
+#endif
 	int err;
 
 	ASSERT(!(sb->s_flags & SB_RDONLY));
 
+	bh = apfs_getblk(sb, nxi->nx_bno);
+	if (!bh) {
+		apfs_err(sb, "failed to map new checkpoint superblock");
+		return -EIO;
+	}
+	obj->o_xid = cpu_to_le64(nxi->nx_xid);
+	apfs_obj_set_csum(sb, obj);
+	memcpy(bh->b_data, obj, sb->s_blocksize);
+
 	err = filemap_write_and_wait(bdev_map);
 	if (err)
-		return err;
+		goto out;
 
-	apfs_obj_set_csum(sb, obj);
 	mark_buffer_dirty(bh);
 	err = sync_dirty_buffer(bh);
 	if (err)
-		return err;
+		goto out;
 
-	return filemap_write_and_wait(bdev_map);
+	err = filemap_write_and_wait(bdev_map);
+out:
+	brelse(bh);
+	bh = NULL;
+	return err;
 }
 
 /**
@@ -423,6 +91,230 @@ static bool apfs_transaction_has_room(struct super_block *sb, struct apfs_max_op
 	return max_blks < APFS_SM(sb)->sm_free_count;
 }
 
+/**
+ * apfs_read_single_ephemeral_object - Read a single ephemeral object to memory
+ * @sb:		filesystem superblock
+ * @map:	checkpoint mapping for the object
+ *
+ * Returns 0 on success or a negative error code in case of failure.
+ */
+static int apfs_read_single_ephemeral_object(struct super_block *sb, struct apfs_checkpoint_mapping *map)
+{
+	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
+	struct apfs_ephemeral_object_info *list = NULL;
+	struct buffer_head *bh = NULL;
+	char *object = NULL;
+	int count;
+	u32 size;
+	u64 bno, oid;
+	int err, i;
+
+	list = nxi->nx_eph_list;
+	count = nxi->nx_eph_count;
+	if (count >= APFS_EPHEMERAL_LIST_LIMIT) {
+		apfs_err(sb, "too many ephemeral objects?");
+		return -EOPNOTSUPP;
+	}
+
+	bno = le64_to_cpu(map->cpm_paddr);
+	oid = le64_to_cpu(map->cpm_oid);
+	size = le32_to_cpu(map->cpm_size);
+	if (size > sb->s_blocksize << 1) {
+		/*
+		 * No reason not to support bigger objects, but there has to be
+		 * a limit somewhere and this is all I've seen so far.
+		 */
+		apfs_warn(sb, "ephemeral object has more than 2 blocks");
+		return -EOPNOTSUPP;
+	}
+	if (!size || (size & (sb->s_blocksize - 1))) {
+		apfs_err(sb, "invalid object size (0x%x)", size);
+		return -EFSCORRUPTED;
+	}
+	object = kmalloc(size, GFP_KERNEL);
+	if (!object)
+		return -ENOMEM;
+
+	for (i = 0; i < size >> sb->s_blocksize_bits; ++i) {
+		bh = apfs_sb_bread(sb, bno + i);
+		if (!bh) {
+			apfs_err(sb, "failed to read ephemeral block");
+			err = -EIO;
+			goto fail;
+		}
+		memcpy(object + (i << sb->s_blocksize_bits), bh->b_data, sb->s_blocksize);
+		brelse(bh);
+		bh = NULL;
+	}
+
+	/*
+	 * The official reference requires that we always verify ephemeral
+	 * checksums on mount, so do it even if the user didn't ask. We should
+	 * actually try to mount an older checkpoint when this fails (TODO),
+	 * which I guess means that the official driver writes all checkpoint
+	 * blocks at once, instead of leaving the superblock for last like we
+	 * do.
+	 */
+	if (!apfs_multiblock_verify_csum(object, size)) {
+		apfs_err(sb, "bad checksum for ephemeral object 0x%llx", oid);
+		err = -EFSBADCRC;
+		goto fail;
+	}
+
+	list[count].oid = oid;
+	list[count].size = size;
+	list[count].object = object;
+	object = NULL;
+	nxi->nx_eph_count = count + 1;
+	return 0;
+
+fail:
+	kfree(object);
+	object = NULL;
+	return err;
+}
+
+/**
+ * apfs_read_single_cpm_block - Read all ephemeral objects in a cpm block
+ * @sb:		filesystem superblock
+ * @cpm_bno:	block number for the cpm block
+ *
+ * Returns 0 on success or a negative error code in case of failure.
+ */
+static int apfs_read_single_cpm_block(struct super_block *sb, u64 cpm_bno)
+{
+	struct buffer_head *bh = NULL;
+	struct apfs_checkpoint_map_phys *cpm = NULL;
+	u32 map_count;
+	int err, i;
+
+	bh = apfs_sb_bread(sb, cpm_bno);
+	if (!bh) {
+		apfs_err(sb, "failed to read cpm block");
+		return -EIO;
+	}
+	if (!apfs_obj_verify_csum(sb, bh)) {
+		/*
+		 * The reference seems to imply that we need to check these on
+		 * mount, and retry an older checkpoint on failure (TODO).
+		 */
+		apfs_err(sb, "bad checksum for cpm block at 0x%llx", cpm_bno);
+		err = -EFSBADCRC;
+		goto out;
+	}
+	cpm = (struct apfs_checkpoint_map_phys *)bh->b_data;
+
+	map_count = le32_to_cpu(cpm->cpm_count);
+	if (map_count > apfs_max_maps_per_block(sb)) {
+		apfs_err(sb, "block has too many maps (%d)", map_count);
+		err = -EFSCORRUPTED;
+		goto out;
+	}
+
+	for (i = 0; i < map_count; ++i) {
+		err = apfs_read_single_ephemeral_object(sb, &cpm->cpm_map[i]);
+		if (err) {
+			apfs_err(sb, "failed to read ephemeral object %u", i);
+			goto out;
+		}
+	}
+
+out:
+	brelse(bh);
+	cpm = NULL;
+	return err;
+}
+
+/**
+ * apfs_read_ephemeral_objects - Read all ephemeral objects to memory
+ * @sb:	superblock structure
+ *
+ * Returns 0 on success or a negative error code in case of failure.
+ */
+static int apfs_read_ephemeral_objects(struct super_block *sb)
+{
+	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
+	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
+	u64 desc_base;
+	u32 desc_index, desc_blks, desc_len, i;
+	int err;
+
+	if (nxi->nx_eph_list) {
+		apfs_alert(sb, "attempt to reread ephemeral object list");
+		return -EFSCORRUPTED;
+	}
+	nxi->nx_eph_list = kzalloc(APFS_EPHEMERAL_LIST_SIZE, GFP_KERNEL);
+	if (!nxi->nx_eph_list)
+		return -ENOMEM;
+	nxi->nx_eph_count = 0;
+
+	desc_base = le64_to_cpu(raw_sb->nx_xp_desc_base);
+	desc_index = le32_to_cpu(raw_sb->nx_xp_desc_index);
+	desc_blks = le32_to_cpu(raw_sb->nx_xp_desc_blocks);
+	desc_len = le32_to_cpu(raw_sb->nx_xp_desc_len);
+
+	/* Last block in the area is superblock; the rest are mapping blocks */
+	for (i = 0; i < desc_len - 1; ++i) {
+		u64 cpm_bno = desc_base + (desc_index + i) % desc_blks;
+
+		err = apfs_read_single_cpm_block(sb, cpm_bno);
+		if (err) {
+			apfs_err(sb, "failed to read cpm block %u", i);
+			return err;
+		}
+	}
+	return 0;
+}
+
+static void apfs_trans_commit_work(struct work_struct *work)
+{
+	struct super_block *sb = NULL;
+	struct apfs_nxsb_info *nxi = NULL;
+	struct apfs_nx_transaction *trans = NULL;
+	int err;
+
+	trans = container_of(to_delayed_work(work), struct apfs_nx_transaction, t_work);
+	nxi = container_of(trans, struct apfs_nxsb_info, nx_transaction);
+	sb = trans->t_work_sb;
+
+	/*
+	 * If sb is set then the transaction already started, there is no need
+	 * for apfs_transaction_start() here. It would be cleaner to call it
+	 * anyway (and check in there if sb is set), but maxops is a problem
+	 * because we don't need any space. I really need to rethink that stuff
+	 * (TODO).
+	 */
+	down_write(&nxi->nx_big_sem);
+	mutex_lock(&nxs_mutex);
+	if (!sb || sb->s_flags & SB_RDONLY) {
+		/* The commit already took place, or there was an abort */
+		mutex_unlock(&nxs_mutex);
+		up_write(&nxi->nx_big_sem);
+		return;
+	}
+
+	trans->t_state |= APFS_NX_TRANS_FORCE_COMMIT;
+	err = apfs_transaction_commit(sb);
+	if (err) {
+		apfs_err(sb, "queued commit failed (err:%d)", err);
+		apfs_transaction_abort(sb);
+	}
+}
+
+/**
+ * apfs_transaction_init - Initialize the transaction struct for the container
+ * @trans: the transaction structure
+ */
+void apfs_transaction_init(struct apfs_nx_transaction *trans)
+{
+	trans->t_state = 0;
+	INIT_DELAYED_WORK(&trans->t_work, apfs_trans_commit_work);
+	INIT_LIST_HEAD(&trans->t_inodes);
+	INIT_LIST_HEAD(&trans->t_buffers);
+	trans->t_buffers_count = 0;
+	trans->t_starts_count = 0;
+}
+
 /**
  * apfs_transaction_start - Begin a new transaction
  * @sb:		superblock structure
@@ -433,9 +325,7 @@ static bool apfs_transaction_has_room(struct super_block *sb, struct apfs_max_op
  */
 int apfs_transaction_start(struct super_block *sb, struct apfs_max_ops maxops)
 {
-	struct apfs_sb_info *sbi = APFS_SB(sb);
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_vol_transaction *vol_trans = &sbi->s_transaction;
 	struct apfs_nx_transaction *nx_trans = &nxi->nx_transaction;
 	int err;
 
@@ -449,23 +339,23 @@ int apfs_transaction_start(struct super_block *sb, struct apfs_max_ops maxops)
 		return -EROFS;
 	}
 
-	/* TODO: rethink this now that transactions shouldn't fail */
-	if (!nx_trans->t_old_msb) {
-		/* Backup the old superblock buffers in case the transaction fails */
-		nx_trans->t_old_msb = nxi->nx_object.o_bh;
-		get_bh(nx_trans->t_old_msb);
-
-		++nxi->nx_xid;
-		INIT_LIST_HEAD(&nx_trans->t_inodes);
-		INIT_LIST_HEAD(&nx_trans->t_buffers);
-		nx_trans->t_buffers_count = 0;
-		nx_trans->t_starts_count = 0;
-
-		err = apfs_checkpoint_start(sb);
+	/*
+	 * Ephemeral objects are read only once, kept in memory, and committed
+	 * to disk along with each transaction.
+	 */
+	if (!nxi->nx_eph_list) {
+		err = apfs_read_ephemeral_objects(sb);
 		if (err) {
-			apfs_err(sb, "failed to start a new checkpoint");
-			goto fail;
+			mutex_unlock(&nxs_mutex);
+			up_write(&nxi->nx_big_sem);
+			apfs_err(sb, "failed to read the ephemeral objects");
+			return err;
 		}
+	}
+
+	if (nx_trans->t_starts_count == 0) {
+		++nxi->nx_xid;
+		nxi->nx_raw->nx_next_xid = cpu_to_le64(nxi->nx_xid + 1);
 
 		err = apfs_read_spaceman(sb);
 		if (err) {
@@ -486,16 +376,7 @@ int apfs_transaction_start(struct super_block *sb, struct apfs_max_ops maxops)
 		return -ENOSPC;
 	}
 
-	if (!vol_trans->t_old_vsb) {
-		vol_trans->t_old_vsb = sbi->s_vobject.o_bh;
-		get_bh(vol_trans->t_old_vsb);
-
-		/* Backup the old tree roots; the node struct issues make this ugly */
-		vol_trans->t_old_cat_root = *sbi->s_cat_root;
-		get_bh(vol_trans->t_old_cat_root.object.o_bh);
-		vol_trans->t_old_omap_root = *sbi->s_omap->omap_root;
-		get_bh(vol_trans->t_old_omap_root.object.o_bh);
-
+	if (nx_trans->t_starts_count == 0) {
 		err = apfs_map_volume_super(sb, true /* write */);
 		if (err) {
 			apfs_err(sb, "CoW failed for volume super");
@@ -523,42 +404,6 @@ fail:
 	return err;
 }
 
-/**
- * apfs_end_buffer_write_sync - Clean up a buffer head just synced to disk
- * @bh:		the buffer head to clean
- * @uptodate:	has the write succeeded?
- */
-static void apfs_end_buffer_write_sync(struct buffer_head *bh, int uptodate)
-{
-	struct page *page = NULL;
-	bool must_unlock, is_metadata;
-
-	page = bh->b_page;
-	get_page(page);
-
-	is_metadata = buffer_csum(bh);
-	clear_buffer_csum(bh);
-	end_buffer_write_sync(bh, uptodate);
-	bh = NULL;
-
-	/* Future writes to mmapped areas should fault for CoW */
-	must_unlock = trylock_page(page);
-	page_mkclean(page);
-
-	/* XXX: otherwise, the page cache fills up and crashes the machine */
-	if (!is_metadata) {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
-		try_to_free_buffers(page_folio(page));
-#else
-		try_to_free_buffers(page);
-#endif
-	}
-
-	if (must_unlock)
-		unlock_page(page);
-	put_page(page);
-}
-
 /**
  * apfs_transaction_flush_all_inodes - Flush inode metadata to the buffer heads
  * @sb: superblock structure
@@ -574,7 +419,6 @@ int apfs_transaction_flush_all_inodes(struct super_block *sb)
 	int err = 0, curr_err;
 
 	ASSERT(!(sb->s_flags & SB_RDONLY));
-	ASSERT(nx_trans->t_old_msb);
 
 	while (!list_empty(&nx_trans->t_inodes)) {
 		struct apfs_inode_info *ai = NULL;
@@ -599,16 +443,16 @@ int apfs_transaction_flush_all_inodes(struct super_block *sb)
 		mutex_unlock(&nxs_mutex);
 		up_write(&nxi->nx_big_sem);
 
-		/* Unlocked, so it may call ->evict_inode() */
+		/* Unlocked, so it may call evict() and wait for writeback */
 		iput(inode);
 
 		down_write(&nxi->nx_big_sem);
 		mutex_lock(&nxs_mutex);
 		nx_trans->t_state = 0;
 
-		/* Transaction aborted by ->evict_inode(), error code is lost */
+		/* Transaction aborted during writeback, error code is lost */
 		if (sb->s_flags & SB_RDONLY) {
-			apfs_err(sb, "abort during inode eviction");
+			apfs_err(sb, "abort during inode writeback");
 			return -EROFS;
 		}
 	}
@@ -616,6 +460,173 @@ int apfs_transaction_flush_all_inodes(struct super_block *sb)
 	return err;
 }
 
+/**
+ * apfs_write_single_ephemeral_object - Write a single ephemeral object to bh's
+ * @sb:		filesystem superblock
+ * @obj_raw:	contents of the object
+ * @map:	checkpoint mapping for the object, already updated
+ *
+ * Returns 0 on success or a negative error code in case of failure.
+ */
+static int apfs_write_single_ephemeral_object(struct super_block *sb, struct apfs_obj_phys *obj_raw, const struct apfs_checkpoint_mapping *map)
+{
+	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
+	struct buffer_head *bh = NULL;
+	u64 bno;
+	u32 size;
+	int err, i;
+
+	bno = le64_to_cpu(map->cpm_paddr);
+	size = le32_to_cpu(map->cpm_size);
+	obj_raw->o_xid = cpu_to_le64(nxi->nx_xid);
+	apfs_multiblock_set_csum((char *)obj_raw, size);
+
+	for (i = 0; i < size >> sb->s_blocksize_bits; ++i) {
+		bh = apfs_getblk(sb, bno + i);
+		if (!bh) {
+			apfs_err(sb, "failed to map ephemeral block");
+			return -EIO;
+		}
+		err = apfs_transaction_join(sb, bh);
+		if (err) {
+			brelse(bh);
+			bh = NULL;
+			return err;
+		}
+		memcpy(bh->b_data, (char *)obj_raw + (i << sb->s_blocksize_bits), sb->s_blocksize);
+		brelse(bh);
+		bh = NULL;
+	}
+	return 0;
+}
+
+/**
+ * apfs_write_ephemeral_objects - Write all ephemeral objects to bh's
+ * @sb: filesystem superblock
+ *
+ * Returns 0 on sucess, or a negative error code in case of failure.
+ */
+static int apfs_write_ephemeral_objects(struct super_block *sb)
+{
+	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
+	struct apfs_nx_superblock *raw_sb = nxi->nx_raw;
+	struct apfs_checkpoint_map_phys *cpm = NULL;
+	struct buffer_head *cpm_bh = NULL;
+	struct apfs_ephemeral_object_info *eph_info = NULL;
+	u64 cpm_bno;
+	u64 desc_base, data_base;
+	u32 desc_index, desc_blks, desc_len, desc_next;
+	u32 data_index, data_blks, data_len, data_next;
+	u32 desc_limit, data_limit;
+	u32 obj_blkcnt;
+	int err, i, cpm_start;
+
+	if (!nxi->nx_eph_list) {
+		apfs_alert(sb, "missing ephemeral object list");
+		return -EFSCORRUPTED;
+	}
+
+	desc_next = le32_to_cpu(raw_sb->nx_xp_desc_next);
+	desc_base = le64_to_cpu(raw_sb->nx_xp_desc_base);
+	desc_index = desc_next; /* New checkpoint */
+	desc_blks = le32_to_cpu(raw_sb->nx_xp_desc_blocks);
+	desc_len = 0; /* For now */
+
+	data_next = le32_to_cpu(raw_sb->nx_xp_data_next);
+	data_base = le64_to_cpu(raw_sb->nx_xp_data_base);
+	data_index = data_next; /* New checkpoint */
+	data_blks = le32_to_cpu(raw_sb->nx_xp_data_blocks);
+	data_len = 0; /* For now */
+
+	/*
+	 * The reference doesn't mention anything about this, but I need to
+	 * put some sort of a limit or else the rings could wrap around and
+	 * corrupt themselves.
+	 */
+	desc_limit = desc_blks >> 2;
+	data_limit = data_blks >> 2;
+
+	for (i = 0; i < nxi->nx_eph_count; ++i) {
+		if (data_len == data_limit) {
+			apfs_err(sb, "too many checkpoint data blocks");
+			return -EFSCORRUPTED;
+		}
+
+		if (!cpm) {
+			cpm_start = i;
+			if (desc_len == desc_limit) {
+				apfs_err(sb, "too many checkpoint descriptor blocks");
+				return -EFSCORRUPTED;
+			}
+			cpm_bno = desc_base + (desc_index + desc_len) % desc_blks;
+			err = apfs_create_cpm_block(sb, cpm_bno, &cpm_bh);
+			if (err) {
+				apfs_err(sb, "failed to create cpm block");
+				return err;
+			}
+			cpm = (void *)cpm_bh->b_data;
+			desc_len += 1;
+		}
+
+		eph_info = &nxi->nx_eph_list[i];
+		data_next = (data_index + data_len) % data_blks;
+		obj_blkcnt = eph_info->size >> sb->s_blocksize_bits;
+		if (obj_blkcnt > data_blks - data_next) {
+			/*
+			 * This multiblock object does not fit in what's left
+			 * of the ring buffer, so move it to the beginning and
+			 * leave some empty blocks.
+			 */
+			data_len += data_blks - data_next;
+			data_next = 0;
+		}
+
+		err = apfs_create_cpoint_map(sb, cpm, eph_info->object, data_base + data_next, eph_info->size);
+		if (err) {
+			if (err == -ENOSPC)
+				cpm->cpm_flags = 0; /* No longer the last */
+			brelse(cpm_bh);
+			cpm = NULL;
+			cpm_bh = NULL;
+			if (err == -ENOSPC) {
+				--i;
+				continue;
+			}
+			apfs_err(sb, "failed to create cpm map %d", i);
+			return err;
+		}
+		err = apfs_write_single_ephemeral_object(sb, eph_info->object, &cpm->cpm_map[i - cpm_start]);
+		if (err) {
+			brelse(cpm_bh);
+			cpm = NULL;
+			cpm_bh = NULL;
+			apfs_err(sb, "failed to write ephemeral object %d", i);
+			return err;
+		}
+		data_len += obj_blkcnt;
+	}
+
+	/*
+	 * The checkpoint superblock can't be set until the very end of the
+	 * transaction commit, but allocate its block here already.
+	 */
+	nxi->nx_bno = desc_base + (desc_index + desc_len) % desc_blks;
+	desc_len += 1;
+
+	desc_next = (desc_index + desc_len) % desc_blks;
+	data_next = (data_index + data_len) % data_blks;
+
+	raw_sb->nx_xp_desc_next = cpu_to_le32(desc_next);
+	raw_sb->nx_xp_desc_index = cpu_to_le32(desc_index);
+	raw_sb->nx_xp_desc_len = cpu_to_le32(desc_len);
+
+	raw_sb->nx_xp_data_next = cpu_to_le32(data_next);
+	raw_sb->nx_xp_data_index = cpu_to_le32(data_index);
+	raw_sb->nx_xp_data_len = cpu_to_le32(data_len);
+
+	return 0;
+}
+
 /**
  * apfs_transaction_commit_nx - Definitely commit the current transaction
  * @sb: superblock structure
@@ -624,13 +635,12 @@ static int apfs_transaction_commit_nx(struct super_block *sb)
 {
 	struct apfs_spaceman *sm = APFS_SM(sb);
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
-	struct apfs_sb_info *sbi;
 	struct apfs_nx_transaction *nx_trans = &nxi->nx_transaction;
 	struct apfs_bh_info *bhi, *tmp;
 	int err = 0;
+	u32 bmap_idx;
 
 	ASSERT(!(sb->s_flags & SB_RDONLY));
-	ASSERT(nx_trans->t_old_msb);
 
 	/* Before committing the bhs, write all inode metadata to them */
 	err = apfs_transaction_flush_all_inodes(sb);
@@ -641,7 +651,8 @@ static int apfs_transaction_commit_nx(struct super_block *sb)
 
 	/*
 	 * Now that nothing else will be freed, flush the last update to the
-	 * free queues so that it can be committed to disk with the other bhs
+	 * free queues so that it can be committed to disk along with all the
+	 * ephemeral objects.
 	 */
 	if (sm->sm_free_cache_base) {
 		err = apfs_free_queue_insert_nocache(sb, sm->sm_free_cache_base, sm->sm_free_cache_blkcnt);
@@ -651,8 +662,11 @@ static int apfs_transaction_commit_nx(struct super_block *sb)
 		}
 		sm->sm_free_cache_base = sm->sm_free_cache_blkcnt = 0;
 	}
+	err = apfs_write_ephemeral_objects(sb);
+	if (err)
+		return err;
 
-	list_for_each_entry_safe(bhi, tmp, &nx_trans->t_buffers, list) {
+	list_for_each_entry(bhi, &nx_trans->t_buffers, list) {
 		struct buffer_head *bh = bhi->bh;
 
 		ASSERT(buffer_trans(bh));
@@ -660,6 +674,25 @@ static int apfs_transaction_commit_nx(struct super_block *sb)
 		if (buffer_csum(bh))
 			apfs_obj_set_csum(sb, (void *)bh->b_data);
 
+		clear_buffer_dirty(bh);
+		get_bh(bh);
+		lock_buffer(bh);
+		bh->b_end_io = end_buffer_write_sync;
+		apfs_submit_bh(REQ_OP_WRITE, REQ_SYNC, bh);
+	}
+	list_for_each_entry_safe(bhi, tmp, &nx_trans->t_buffers, list) {
+		struct buffer_head *bh = bhi->bh;
+		struct page *page = NULL;
+		bool is_metadata;
+
+		ASSERT(buffer_trans(bh));
+
+		wait_on_buffer(bh);
+		if (!buffer_uptodate(bh)) {
+			apfs_err(sb, "failed to write some blocks");
+			return -EIO;
+		}
+
 		list_del(&bhi->list);
 		clear_buffer_trans(bh);
 		nx_trans->t_buffers_count--;
@@ -669,10 +702,27 @@ static int apfs_transaction_commit_nx(struct super_block *sb)
 		kfree(bhi);
 		bhi = NULL;
 
-		clear_buffer_dirty(bh);
-		lock_buffer(bh);
-		bh->b_end_io = apfs_end_buffer_write_sync;
-		apfs_submit_bh(REQ_OP_WRITE, REQ_SYNC, bh);
+		page = bh->b_page;
+		get_page(page);
+
+		is_metadata = buffer_csum(bh);
+		clear_buffer_csum(bh);
+		put_bh(bh);
+		bh = NULL;
+
+		/* Future writes to mmapped areas should fault for CoW */
+		lock_page(page);
+		page_mkclean(page);
+		/* XXX: otherwise, the page cache fills up and crashes the machine */
+		if (!is_metadata) {
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
+			try_to_free_buffers(page_folio(page));
+#else
+			try_to_free_buffers(page);
+#endif
+		}
+		unlock_page(page);
+		put_page(page);
 	}
 	err = apfs_checkpoint_end(sb);
 	if (err) {
@@ -680,31 +730,14 @@ static int apfs_transaction_commit_nx(struct super_block *sb)
 		return err;
 	}
 
-	/* Success: forget the old container and volume superblocks */
-	brelse(nx_trans->t_old_msb);
-	nx_trans->t_old_msb = NULL;
-
-	list_for_each_entry(sbi, &nxi->vol_list, list) {
-		struct apfs_vol_transaction *vol_trans = &sbi->s_transaction;
-		if (!vol_trans->t_old_vsb)
-			continue;
-
-		brelse(vol_trans->t_old_vsb);
-		vol_trans->t_old_vsb = NULL;
-
-		/* XXX: forget the buffers for the b-tree roots */
-		vol_trans->t_old_omap_root.object.data = NULL;
-		brelse(vol_trans->t_old_omap_root.object.o_bh);
-		vol_trans->t_old_omap_root.object.o_bh = NULL;
-		vol_trans->t_old_cat_root.object.data = NULL;
-		brelse(vol_trans->t_old_cat_root.object.o_bh);
-		vol_trans->t_old_cat_root.object.o_bh = NULL;
+	for (bmap_idx = 0; bmap_idx < APFS_SM(sb)->sm_ip_bmaps_count; ++bmap_idx) {
+		brelse(APFS_SM(sb)->sm_ip_bmaps[bmap_idx]);
+		APFS_SM(sb)->sm_ip_bmaps[bmap_idx] = NULL;
 	}
+	APFS_SM(sb)->sm_raw = NULL;
 
-	brelse(APFS_SM(sb)->sm_ip);
-	APFS_SM(sb)->sm_ip = NULL;
-	brelse(APFS_SM(sb)->sm_bh);
-	APFS_SM(sb)->sm_bh = NULL;
+	nx_trans->t_starts_count = 0;
+	nx_trans->t_buffers_count = 0;
 	return 0;
 }
 
@@ -723,7 +756,7 @@ static bool apfs_transaction_need_commit(struct super_block *sb)
 		return false;
 	}
 
-	/* Avoid nested commits on ->evict_inode() */
+	/* Avoid nested commits on inode writeback */
 	if (nx_trans->t_state & APFS_NX_TRANS_COMMITTING)
 		return false;
 
@@ -736,9 +769,9 @@ static bool apfs_transaction_need_commit(struct super_block *sb)
 		struct apfs_spaceman_phys *sm_raw = sm->sm_raw;
 		struct apfs_spaceman_free_queue *fq_ip = &sm_raw->sm_fq[APFS_SFQ_IP];
 		struct apfs_spaceman_free_queue *fq_main = &sm_raw->sm_fq[APFS_SFQ_MAIN];
-		int buffers_max = TRANSACTION_BUFFERS_MAX;
-		int starts_max = TRANSACTION_STARTS_MAX;
-		int mq_max = TRANSACTION_MAIN_QUEUE_MAX;
+		int buffers_max = nxi->nx_trans_buffers_max;
+		int starts_max = APFS_TRANS_STARTS_MAX;
+		int mq_max = APFS_TRANS_MAIN_QUEUE_MAX;
 
 		/*
 		 * Try to avoid committing halfway through a data block write,
@@ -751,7 +784,7 @@ static bool apfs_transaction_need_commit(struct super_block *sb)
 			mq_max += 20;
 		}
 
-		if(nx_trans->t_buffers_count > buffers_max)
+		if (nx_trans->t_buffers_count > buffers_max)
 			return true;
 		if (nx_trans->t_starts_count > starts_max)
 			return true;
@@ -761,11 +794,11 @@ static bool apfs_transaction_need_commit(struct super_block *sb)
 		 * exactly 3 times. Don't allow large transactions if we can't
 		 * be sure the bitmap changes will all fit.
 		 */
-		if(le64_to_cpu(fq_ip->sfq_count) * 3 > le64_to_cpu(sm_raw->sm_ip_block_count))
+		if (le64_to_cpu(fq_ip->sfq_count) * 3 > le64_to_cpu(sm_raw->sm_ip_block_count))
 			return true;
 
 		/* Don't let the main queue get too full either */
-		if(le64_to_cpu(fq_main->sfq_count) > mq_max)
+		if (le64_to_cpu(fq_main->sfq_count) > mq_max)
 			return true;
 	}
 
@@ -783,14 +816,22 @@ static bool apfs_transaction_need_commit(struct super_block *sb)
 int apfs_transaction_commit(struct super_block *sb)
 {
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
+	struct apfs_nx_transaction *trans = NULL;
 	int err = 0;
 
+	trans = &nxi->nx_transaction;
+
 	if (apfs_transaction_need_commit(sb)) {
 		err = apfs_transaction_commit_nx(sb);
 		if (err) {
 			apfs_err(sb, "transaction commit failed");
 			return err;
 		}
+		trans->t_work_sb = NULL;
+		cancel_delayed_work(&trans->t_work);
+	} else {
+		trans->t_work_sb = sb;
+		mod_delayed_work(system_wq, &trans->t_work, msecs_to_jiffies(100));
 	}
 
 	mutex_unlock(&nxs_mutex);
@@ -810,7 +851,6 @@ void apfs_inode_join_transaction(struct super_block *sb, struct inode *inode)
 	struct apfs_inode_info *ai = APFS_I(inode);
 
 	ASSERT(!(sb->s_flags & SB_RDONLY));
-	ASSERT(nx_trans->t_old_msb);
 	lockdep_assert_held_write(&nxi->nx_big_sem);
 
 	if (!list_empty(&ai->i_list)) /* Already in the transaction */
@@ -832,7 +872,6 @@ int apfs_transaction_join(struct super_block *sb, struct buffer_head *bh)
 	struct apfs_bh_info *bhi;
 
 	ASSERT(!(sb->s_flags & SB_RDONLY));
-	ASSERT(nx_trans->t_old_msb);
 	lockdep_assert_held_write(&nxi->nx_big_sem);
 
 	if (buffer_trans(bh)) /* Already part of the only transaction */
@@ -845,7 +884,7 @@ int apfs_transaction_join(struct super_block *sb, struct buffer_head *bh)
 	get_bh(bh);
 	bhi->bh = bh;
 	list_add(&bhi->list, &nx_trans->t_buffers);
-	nx_trans->t_buffers_count ++;
+	nx_trans->t_buffers_count++;
 
 	set_buffer_trans(bh);
 	bh->b_private = bhi;
@@ -878,15 +917,15 @@ static void apfs_force_readonly(struct apfs_nxsb_info *nxi)
  */
 void apfs_transaction_abort(struct super_block *sb)
 {
-	struct apfs_sb_info *sbi;
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
 	struct apfs_nx_transaction *nx_trans = &nxi->nx_transaction;
 	struct apfs_bh_info *bhi, *tmp;
 	struct apfs_inode_info *ai, *ai_tmp;
+	struct apfs_spaceman *sm = NULL;
+	u32 bmap_idx;
 
 	if (sb->s_flags & SB_RDONLY) {
 		/* Transaction already aborted, do nothing */
-		ASSERT(!nx_trans->t_old_msb);
 		ASSERT(list_empty(&nx_trans->t_inodes));
 		ASSERT(list_empty(&nx_trans->t_buffers));
 		mutex_unlock(&nxs_mutex);
@@ -894,7 +933,6 @@ void apfs_transaction_abort(struct super_block *sb)
 		return;
 	}
 
-	ASSERT(nx_trans->t_old_msb);
 	nx_trans->t_state = 0;
 	apfs_err(sb, "aborting transaction");
 
@@ -913,43 +951,15 @@ void apfs_transaction_abort(struct super_block *sb)
 		kfree(bhi);
 	}
 
-	/* Restore the old container superblock */
-	brelse(nxi->nx_object.o_bh);
-	nxi->nx_object.o_bh = nx_trans->t_old_msb;
-	nxi->nx_object.data = nxi->nx_object.o_bh->b_data;
-	nxi->nx_object.block_nr = nx_trans->t_old_msb->b_blocknr;
-	nxi->nx_raw = (void *)nx_trans->t_old_msb->b_data;
-	nx_trans->t_old_msb = NULL;
-
-	list_for_each_entry(sbi, &nxi->vol_list, list) {
-		struct apfs_vol_transaction *vol_trans = &sbi->s_transaction;
-		if (!vol_trans->t_old_vsb)
-			continue;
-
-		/* Restore volume state for all volumes */
-		brelse(sbi->s_vobject.o_bh);
-		sbi->s_vobject.o_bh = vol_trans->t_old_vsb;
-		sbi->s_vobject.data = sbi->s_vobject.o_bh->b_data;
-		sbi->s_vobject.block_nr = vol_trans->t_old_vsb->b_blocknr;
-		sbi->s_vsb_raw = (void *)vol_trans->t_old_vsb->b_data;
-		vol_trans->t_old_vsb = NULL;
-
-		/* XXX: restore the old b-tree root nodes */
-		brelse(sbi->s_omap->omap_root->object.o_bh);
-		*(sbi->s_omap->omap_root) = vol_trans->t_old_omap_root;
-		vol_trans->t_old_omap_root.object.o_bh = NULL;
-		vol_trans->t_old_omap_root.object.data = NULL;
-		brelse(sbi->s_cat_root->object.o_bh);
-		*(sbi->s_cat_root) = vol_trans->t_old_cat_root;
-		vol_trans->t_old_cat_root.object.o_bh = NULL;
-		vol_trans->t_old_cat_root.object.data = NULL;
+	sm = APFS_SM(sb);
+	if (sm) {
+		for (bmap_idx = 0; bmap_idx < sm->sm_ip_bmaps_count; ++bmap_idx) {
+			brelse(sm->sm_ip_bmaps[bmap_idx]);
+			sm->sm_ip_bmaps[bmap_idx] = NULL;
+		}
+		APFS_SM(sb)->sm_raw = NULL;
 	}
 
-	brelse(APFS_SM(sb)->sm_ip);
-	APFS_SM(sb)->sm_ip = NULL;
-	brelse(APFS_SM(sb)->sm_bh);
-	APFS_SM(sb)->sm_bh = NULL;
-
 	/*
 	 * It's not possible to undo in-memory changes from old operations in
 	 * the aborted transaction. To avoid corruption, never write again.
@@ -959,7 +969,6 @@ void apfs_transaction_abort(struct super_block *sb)
 	mutex_unlock(&nxs_mutex);
 	up_write(&nxi->nx_big_sem);
 
-	/* ->evict_inode() will just fail if it starts a new transaction */
 	list_for_each_entry_safe(ai, ai_tmp, &nx_trans->t_inodes, i_list) {
 		list_del_init(&ai->i_list);
 		iput(&ai->vfs_inode);
